from datetime import (
    datetime,
    timedelta,
)
from typing import (
    Any,
    Callable,
    Iterable,
    Optional,
    Tuple,
    Union,
)

from .abc import (
    seconds,
    Instrumented_StaticTimeSpanABC,
    StaticTimeSpanABC,
    PeriodicTimePointABC,
    PeriodicTimeSpanABC,
    PeriodicActivityABC,
)


class TimeSpanInitMixin:
    def __init__(
        self,
        *,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
        duration: Optional[seconds] = None,
    ):
        start_nn = start is not None
        end_nn = end is not None
        duration_nn = duration is not None

        if start_nn and end_nn and duration_nn:
            raise ValueError('Cannot specify all three: start, end, duration')

        self.start = start
        self.end = end
        if start_nn and duration_nn:
            self.end = start + timedelta(seconds=duration)
        elif end_nn and duration_nn:
            self.start = end - timedelta(seconds=duration)


class Instrumented_StaticTimeSpan(TimeSpanInitMixin, Instrumented_StaticTimeSpanABC):
    def __bool__(self) -> bool:
        return bool(self.start) and bool(self.end)

    def __hash__(self) -> int:
        return hash((self.start, self.end))
    
    def __eq__(self, other: Any) -> bool:
        return issubclass(other, StaticTimeSpanABC) and (
            self.start == other.start and
            self.end == other.end
        )

    def __lt__(self, other: StaticTimeSpanABC) -> bool:
        self_start, other_start = self.start, other.start
        return not other_start or (self_start and self_start < other_start)

    def __gt__(self, other: StaticTimeSpanABC) -> bool:
        self_start, other_start = self.start, other.start
        return not self_start or (other_start and other_start < self_start)


class PeriodicTimeSpan(PeriodicTimeSpanABC):
    def __init__(self, period_engine: PeriodicTimePointABC, duration: seconds):
        self.period_engine = period_engine
        self.duration = duration

    def is_ongoing(self, moment: datetime) -> bool:
        return bool(self.get_current(moment))

    def get_current(self, moment: datetime) -> Instrumented_StaticTimeSpanABC:
        next = self.period_engine.get_next(moment)
        current = self.period_engine.get_next(moment - timedelta(seconds=self.duration))
        if (not next) or (current and (next > current)):
            return Instrumented_StaticTimeSpan(start=current, duration=self.duration)
        else:
            return Instrumented_StaticTimeSpan()

    def get_next(self, moment: datetime) -> Instrumented_StaticTimeSpanABC:
        return Instrumented_StaticTimeSpan(
            start=self.period_engine.get_next(moment),
            duration=self.duration,
        )

    def get_current_or_next(
        self, moment: datetime, *, return_is_current: bool = False
    ) -> Union[Instrumented_StaticTimeSpanABC, Tuple[Instrumented_StaticTimeSpanABC, Optional[bool]]]:
        if span := self.get_current(moment):
            return (span, True) if return_is_current else span

        if span := self.get_next(moment):
            return (span, False) if return_is_current else span

        span = Instrumented_StaticTimeSpan()
        return (span, None) if return_is_current else span


class PeriodicTimeSpanWithExtraConditions(PeriodicTimeSpan):
    def __init__(
        self,
        period_engine: PeriodicTimePointABC,
        duration: seconds,
        extra_conditions: Iterable[Callable[[Instrumented_StaticTimeSpanABC], bool]] = None,
        extra_conditions_any: bool = False,
    ):
        super().__init__(period_engine, duration)
        self.extra_conditions = extra_conditions
        self.extra_conditions_any = extra_conditions_any

    def extra_conditions_ok(
        self, span: Instrumented_StaticTimeSpanABC
    ) -> bool:
        if not self.extra_conditions:
            return True

        return (
            any(map(lambda ec: ec(span), self.extra_conditions))
            if self.extra_conditions_any else
            all(map(lambda ec: ec(span), self.extra_conditions))
        )

    def get_current(self, moment: datetime) -> Instrumented_StaticTimeSpanABC:
        span = super().get_current(moment)
        if span and self.extra_conditions_ok(span):
            return span
        else:
            return Instrumented_StaticTimeSpan()

    def get_next(
        self, moment: datetime, extra_conditions_max_fails=20
    ) -> Instrumented_StaticTimeSpanABC:
        span = Instrumented_StaticTimeSpan(start=moment, duration=0)
        while True:
            span = super().get_next(span.start)
            if span:
                if not self.extra_conditions_ok(span):
                    if extra_conditions_max_fails is not None:
                        extra_conditions_max_fails -= 1
                        if extra_conditions_max_fails <= 0:
                            return Instrumented_StaticTimeSpan()
                else:
                    break
            else:
                break

        return span


class PeriodicActivity(PeriodicTimeSpan, PeriodicActivityABC):
    def __init__(
        self,
        payload: Any,
        period_engine: PeriodicTimePointABC,
        duration: seconds,
    ):
        super().__init__(period_engine, duration)
        self.payload = payload


class PeriodicActivityWithExtraConditions(PeriodicTimeSpanWithExtraConditions, PeriodicActivityABC):
    def __init__(
        self,
        payload: Any,
        period_engine: PeriodicTimePointABC,
        duration: seconds,
        extra_conditions: Iterable[Callable[[Instrumented_StaticTimeSpanABC], bool]] = None,
        extra_conditions_any: bool = False,
    ):
        super().__init__(period_engine, duration, extra_conditions, extra_conditions_any)
        self.payload = payload


class ResolvedActivity(StaticTimeSpanABC):
    """An activity with known start/end time"""

    def __init__(self, payload: Any, start: datetime, end: datetime):
        self.payload = payload
        self.start = start
        self.end = end

    def __repr__(self):
        return f'<{self.payload!s}: [{self.start!s} ... {self.end!s}]>'


class ManualSchedule:
    def __init__(self, activities: Iterable[PeriodicActivityABC]):
        self.activities = tuple(activities)

    def get_next(self, moment: datetime) -> Tuple[ResolvedActivity]:
        activity__next__s = self._activity__next__s(moment)
        if not activity__next__s:
            return ()
        soonest_span = min(map(lambda an: an[1], activity__next__s))
        return () if not soonest_span else tuple(
            ResolvedActivity(activity.payload, span.start, span.end)
            for activity, span in activity__next__s
            if span.start == soonest_span.start
        )

    def _activity__next__s(
        self, moment: datetime
    ) -> Tuple[Tuple[PeriodicActivityABC, Instrumented_StaticTimeSpanABC]]:
        """Return ``((<activity>, <next time span>), ...)``"""
        return tuple(map(
            lambda a: (a, a.get_next(moment)),
            self.activities
        ))

    def get_current(self, moment: datetime) -> Tuple[ResolvedActivity]:
        activity__current__s = self._activity__current__s(moment)
        return tuple(
            ResolvedActivity(activity.payload, span.start, span.end)
            for activity, span in activity__current__s
            if span  # if is current
        )

    def _activity__current__s(
        self, moment: datetime
    ) -> Tuple[Tuple[PeriodicActivityABC, Instrumented_StaticTimeSpanABC]]:
        """Return ``((<activity>, <current time span>), ...)``"""
        return tuple(map(
            lambda a: (a, a.get_current(moment)),
            self.activities
        ))

    def get_current_or_next(
        self, moment: datetime, *, return_is_current: bool = False
    ) -> Union[Tuple[ResolvedActivity], Tuple[Tuple[ResolvedActivity], bool]]:
        activity__current_or_next__ongoing__s = self._activity__current_or_next__ongoing__s(moment)
        if not activity__current_or_next__ongoing__s:
            return ((), None) if return_is_current else ()
        
        current_s = tuple(
            ResolvedActivity(activity.payload, span.start, span.end)
            for activity, span, ongoing in activity__current_or_next__ongoing__s
            if ongoing
        )
        if current_s:
            return (current_s, True) if return_is_current else current_s

        soonest_span = min(map(lambda acono: acono[1], activity__current_or_next__ongoing__s))
        next_s = tuple(
            ResolvedActivity(activity.payload, span.start, span.end)
            for activity, span, ongoing in activity__current_or_next__ongoing__s
            if ongoing == False and span.start == soonest_span.start
        )
        if next_s:
            return (next_s, False) if return_is_current else next_s

        return ((), None) if return_is_current else ()

    def _activity__current_or_next__ongoing__s(
        self, moment: datetime
    ) -> Tuple[Tuple[PeriodicActivityABC, Instrumented_StaticTimeSpanABC, bool]]:
        """Return ``((<activity>, <current or next time span>, <is it ongoing>), ...)``"""
        def mapper(activity: PeriodicActivityABC):
            span, ongoing = activity.get_current_or_next(moment, return_is_current=True)
            return (activity, span, ongoing)
        return tuple(map(mapper, self.activities))



from .schedule import (
    Instrumented_StaticTimeSpan,
    
    PeriodicTimeSpan,
    PeriodicTimeSpanWithExtraConditions,

    PeriodicActivity,
    PeriodicActivityWithExtraConditions,

    ResolvedActivity,

    ManualSchedule,
)
from .period_engine import (
    CronIterWrapper,
)



from datetime import datetime

from croniter import croniter

from .abc import (
    PeriodicTimePointABC
)


class CronIterWrapper(PeriodicTimePointABC):
    def __init__(self, cron_expression: str, *args, **kwargs):
        self.cron = croniter(cron_expression, *args, **kwargs)

    def get_next(self, moment: datetime) -> datetime:
        return self.cron.get_next(start_time=moment, ret_type=datetime)



import unittest

import arrow

from schedpack import ManualSchedule
from .utils import (
    NeverStartingPeriodicActivity,
    SchoolClass,
    c1,
    c2,
    c3,
    cron,
    even_week,
    odd_week,
    impossible_condition,
)


class TestFilledScheduleWithExtraConditions(unittest.TestCase):
    def setUp(self):
        self.schedule = ManualSchedule([
            SchoolClass('mon,thu c2', cron('mon,thu', c2)),  # mixed

            SchoolClass('mon c1 ew', cron('mon', c1), extra_conditions=[even_week]),
            SchoolClass('mon c3', cron('mon', c3)),

            SchoolClass('tue c1', cron('tue', c1)),
            SchoolClass('tue c2', cron('tue', c2)),
            SchoolClass('tue c3', cron('tue', c3)),

            SchoolClass('wed c1 ow', cron('wed', c1), extra_conditions=[odd_week]),
            SchoolClass('wed c1 ew', cron('wed', c1), extra_conditions=[even_week]),
            SchoolClass('wed c2', cron('wed', c2)),
            SchoolClass('wed c3', cron('wed', c3)),

            SchoolClass('thu c1 ow', cron('thu', c1), extra_conditions=[odd_week]),  
            SchoolClass('thu c1 ew', cron('thu', c1), extra_conditions=[even_week]),                  
            SchoolClass('thu c3', cron('thu', c3)),

            SchoolClass('fri c1', cron('fri', c1)),
            SchoolClass('fri c2', cron('fri', c2)),
            SchoolClass('fry c3', cron('fri', c3)),

            SchoolClass('sat c2', cron('sat', c2)),
            SchoolClass('sat c3', cron('sat', c3)),

            SchoolClass('sun c11', cron('sun', c1)),  # simultaneous
            SchoolClass('sun c12', cron('sun', c1)),
            SchoolClass('sun c2', cron('sun', c2)),
        ])

    # jul 2021 calendar
    # [[ 0,  0,  0,  1,  2,  3,  4],
    #  [ 5,  6,  7,  8,  9, 10, 11],
    #  [12, 13, 14, 15, 16, 17, 18],
    #  [19, 20, 21, 22, 23, 24, 25],
    #  [26, 27, 28, 29, 30, 31,  0]]

    def test_schedule_get_next__when_no_current_and_extra_conditions(self):
        moment = arrow.get('2021-07-01T07:30:00', tzinfo='Europe/Moscow').datetime

        activities = self.schedule.get_next(moment)

        self.assertEqual(len(activities), 1)
        self.assertEqual(activities[0].payload, 'thu c1 ow')
        self.assertEqual(
            activities[0].start,
            arrow.get('2021-07-01T08:00:00', tzinfo='Europe/Moscow').datetime
        )
        self.assertEqual(
            activities[0].end,
            arrow.get('2021-07-01T09:35:00', tzinfo='Europe/Moscow').datetime
        )

    def test_schedule_get_next__when_start_equals_moment(self):
        moment = arrow.get('2021-07-02T08:00:00', tzinfo='Europe/Moscow').datetime

        activities = self.schedule.get_next(moment)

        self.assertEqual(len(activities), 1)
        self.assertEqual(activities[0].payload, 'fri c2')
        self.assertEqual(
            activities[0].start,
            arrow.get('2021-07-02T09:50:00', tzinfo='Europe/Moscow').datetime
        )
        self.assertEqual(
            activities[0].end,
            arrow.get('2021-07-02T11:25:00', tzinfo='Europe/Moscow').datetime
        )

    def test_schedule_get_next__when_no_current_and_many_activities(self):
        moment = arrow.get('2021-07-04T07:30:00', tzinfo='Europe/Moscow').datetime

        activities = self.schedule.get_next(moment)

        self.assertEqual(len(activities), 2)
        self.assertEqual(activities[0].payload, 'sun c11')
        self.assertEqual(
            activities[0].start,
            arrow.get('2021-07-04T08:00:00', tzinfo='Europe/Moscow').datetime
        )
        self.assertEqual(
            activities[0].end,
            arrow.get('2021-07-04T09:35:00', tzinfo='Europe/Moscow').datetime
        )
        self.assertEqual(activities[1].payload, 'sun c12')
        self.assertEqual(
            activities[1].start,
            arrow.get('2021-07-04T08:00:00', tzinfo='Europe/Moscow').datetime
        )
        self.assertEqual(
            activities[1].end,
            arrow.get('2021-07-04T09:35:00', tzinfo='Europe/Moscow').datetime
        )

    def test_schedule_get_next__when_start_equals_moment_and_many_activities(self):
        moment = arrow.get('2021-07-04T08:00:00', tzinfo='Europe/Moscow').datetime

        activities = self.schedule.get_next(moment)

        self.assertEqual(len(activities), 1)
        self.assertEqual(activities[0].payload, 'sun c2')
        self.assertEqual(
            activities[0].start,
            arrow.get('2021-07-04T09:50:00', tzinfo='Europe/Moscow').datetime
        )
        self.assertEqual(
            activities[0].end,
            arrow.get('2021-07-04T11:25:00', tzinfo='Europe/Moscow').datetime
        )

    def test_schedule_get_next__when_moment_equals_end(self):
        moment = arrow.get('2021-07-02T09:35:00', tzinfo='Europe/Moscow').datetime

        activities = self.schedule.get_next(moment)

        self.assertEqual(len(activities), 1)
        self.assertEqual(activities[0].payload, 'fri c2')
        self.assertEqual(
            activities[0].start,
            arrow.get('2021-07-02T09:50:00', tzinfo='Europe/Moscow').datetime
        )
        self.assertEqual(
            activities[0].end,
            arrow.get('2021-07-02T11:25:00', tzinfo='Europe/Moscow').datetime
        )

    def test_schedule_get_current__when_no_current(self):
        moment = arrow.get('2021-07-01T07:30:00', tzinfo='Europe/Moscow').datetime

        activities = self.schedule.get_current(moment)

        self.assertEqual(len(activities), 0)

    def test_schedule_get_current__when_start_equals_moment_and_extra_conditions(self):
        moment = arrow.get('2021-07-01T08:00:00', tzinfo='Europe/Moscow').datetime

        activities = self.schedule.get_current(moment)

        self.assertEqual(len(activities), 1)
        self.assertEqual(activities[0].payload, 'thu c1 ow')
        self.assertEqual(
            activities[0].start,
            arrow.get('2021-07-01T08:00:00', tzinfo='Europe/Moscow').datetime
        )
        self.assertEqual(
            activities[0].end,
            arrow.get('2021-07-01T09:35:00', tzinfo='Europe/Moscow').datetime
        )

    def test_schedule_get_current__when_moment_equals_end(self):
        moment = arrow.get('2021-07-02T09:35:00', tzinfo='Europe/Moscow').datetime

        activities = self.schedule.get_current(moment)

        self.assertEqual(len(activities), 0)

    def test_schedule_get_current__when_there_is_current_and_many_activities(self):
        moment = arrow.get('2021-07-04T08:30:00', tzinfo='Europe/Moscow').datetime

        activities = self.schedule.get_current(moment)

        self.assertEqual(len(activities), 2)
        self.assertEqual(activities[0].payload, 'sun c11')
        self.assertEqual(
            activities[0].start,
            arrow.get('2021-07-04T08:00:00', tzinfo='Europe/Moscow').datetime
        )
        self.assertEqual(
            activities[0].end,
            arrow.get('2021-07-04T09:35:00', tzinfo='Europe/Moscow').datetime
        )
        self.assertEqual(activities[1].payload, 'sun c12')
        self.assertEqual(
            activities[1].start,
            arrow.get('2021-07-04T08:00:00', tzinfo='Europe/Moscow').datetime
        )
        self.assertEqual(
            activities[1].end,
            arrow.get('2021-07-04T09:35:00', tzinfo='Europe/Moscow').datetime
        )

    def test_schedule_get_current_or_next__when_no_current_and_extra_conditions(self):
        moment = arrow.get('2021-07-01T07:30:00', tzinfo='Europe/Moscow').datetime

        activities = self.schedule.get_current_or_next(moment)

        self.assertEqual(len(activities), 1)
        self.assertEqual(activities[0].payload, 'thu c1 ow')
        self.assertEqual(
            activities[0].start,
            arrow.get('2021-07-01T08:00:00', tzinfo='Europe/Moscow').datetime
        )
        self.assertEqual(
            activities[0].end,
            arrow.get('2021-07-01T09:35:00', tzinfo='Europe/Moscow').datetime
        )

    def test_schedule_get_current_or_next__when_there_is_current_and_extra_conditions(self):
        moment = arrow.get('2021-07-01T08:30:00', tzinfo='Europe/Moscow').datetime

        activities = self.schedule.get_current_or_next(moment)

        self.assertEqual(len(activities), 1)
        self.assertEqual(activities[0].payload, 'thu c1 ow')
        self.assertEqual(
            activities[0].start,
            arrow.get('2021-07-01T08:00:00', tzinfo='Europe/Moscow').datetime
        )
        self.assertEqual(
            activities[0].end,
            arrow.get('2021-07-01T09:35:00', tzinfo='Europe/Moscow').datetime
        )

    def test_schedule_get_current_or_next__when_no_current_and_return_is_current(self):
        moment = arrow.get('2021-07-01T07:30:00', tzinfo='Europe/Moscow').datetime

        activities, is_current = self.schedule.get_current_or_next(
            moment, return_is_current=True
        )

        self.assertFalse(is_current)
        self.assertEqual(len(activities), 1)
        self.assertEqual(activities[0].payload, 'thu c1 ow')
        self.assertEqual(
            activities[0].start,
            arrow.get('2021-07-01T08:00:00', tzinfo='Europe/Moscow').datetime
        )
        self.assertEqual(
            activities[0].end,
            arrow.get('2021-07-01T09:35:00', tzinfo='Europe/Moscow').datetime
        )

    def test_schedule_get_current_or_next__when_there_is_current_and_return_is_current(self):
        moment = arrow.get('2021-07-01T08:30:00', tzinfo='Europe/Moscow').datetime

        activities, is_current = self.schedule.get_current_or_next(
            moment, return_is_current=True
        )

        self.assertTrue(is_current)
        self.assertEqual(len(activities), 1)
        self.assertEqual(activities[0].payload, 'thu c1 ow')
        self.assertEqual(
            activities[0].start,
            arrow.get('2021-07-01T08:00:00', tzinfo='Europe/Moscow').datetime
        )
        self.assertEqual(
            activities[0].end,
            arrow.get('2021-07-01T09:35:00', tzinfo='Europe/Moscow').datetime
        )

    def test_schedule_get_current_or_next__when_current_and_return_is_current_and_many_activities(self):
        moment = arrow.get('2021-07-04T08:30:00', tzinfo='Europe/Moscow').datetime

        activities, is_current = self.schedule.get_current_or_next(
            moment, return_is_current=True
        )

        self.assertTrue(is_current)
        self.assertEqual(len(activities), 2)
        self.assertEqual(activities[0].payload, 'sun c11')
        self.assertEqual(
            activities[0].start,
            arrow.get('2021-07-04T08:00:00', tzinfo='Europe/Moscow').datetime
        )
        self.assertEqual(
            activities[0].end,
            arrow.get('2021-07-04T09:35:00', tzinfo='Europe/Moscow').datetime
        )
        self.assertEqual(activities[1].payload, 'sun c12')
        self.assertEqual(
            activities[1].start,
            arrow.get('2021-07-04T08:00:00', tzinfo='Europe/Moscow').datetime
        )
        self.assertEqual(
            activities[1].end,
            arrow.get('2021-07-04T09:35:00', tzinfo='Europe/Moscow').datetime
        )

    def test_schedule_get_current_or_next__when_no_current_and_return_is_current_and_many_activities(self):
        moment = arrow.get('2021-07-04T07:30:00', tzinfo='Europe/Moscow').datetime

        activities, is_current = self.schedule.get_current_or_next(
            moment, return_is_current=True
        )

        self.assertFalse(is_current)
        self.assertEqual(len(activities), 2)
        self.assertEqual(activities[0].payload, 'sun c11')
        self.assertEqual(
            activities[0].start,
            arrow.get('2021-07-04T08:00:00', tzinfo='Europe/Moscow').datetime
        )
        self.assertEqual(
            activities[0].end,
            arrow.get('2021-07-04T09:35:00', tzinfo='Europe/Moscow').datetime
        )
        self.assertEqual(activities[1].payload, 'sun c12')
        self.assertEqual(
            activities[1].start,
            arrow.get('2021-07-04T08:00:00', tzinfo='Europe/Moscow').datetime
        )
        self.assertEqual(
            activities[1].end,
            arrow.get('2021-07-04T09:35:00', tzinfo='Europe/Moscow').datetime
        )


class TestEmptySchedule(unittest.TestCase):
    def setUp(self):
        self.schedule = ManualSchedule([])

    def test_schedule_get_current__when_empty(self):
        moment = arrow.get('2021-07-04T07:30:00', tzinfo='Europe/Moscow').datetime

        activities = self.schedule.get_current(moment)

        self.assertEqual(activities, ())

    def test_schedule_get_next__when_empty(self):
        moment = arrow.get('2021-07-04T07:30:00', tzinfo='Europe/Moscow').datetime

        activities = self.schedule.get_next(moment)

        self.assertEqual(activities, ())

    def test_schedule_get_current_or_next__when_empty(self):
        moment = arrow.get('2021-07-04T07:30:00', tzinfo='Europe/Moscow').datetime

        activities = self.schedule.get_current_or_next(moment)

        self.assertEqual(activities, ())

    def test_schedule_get_current_or_next__when_empty_and_return_is_current(self):
        moment = arrow.get('2021-07-04T07:30:00', tzinfo='Europe/Moscow').datetime

        activities = self.schedule.get_current_or_next(moment, return_is_current=True)

        self.assertEqual(activities, ((), None))


class TestScheduleWithImpossibleExtraConditions(unittest.TestCase):
    def setUp(self):
        self.schedule = ManualSchedule([
            SchoolClass('thu c1', cron('thu', c1), extra_conditions=[impossible_condition]),
            SchoolClass('thu c2', cron('thu', c2)),                
            SchoolClass('thu c3', cron('thu', c3)),

            SchoolClass('fri c11', cron('fri', c1), extra_conditions=[impossible_condition]),
            SchoolClass('fri c12', cron('fri', c1)),
            SchoolClass('fri c2', cron('fri', c2)),
            SchoolClass('fry c3', cron('fri', c3)),

            SchoolClass('sat c2', cron('sat', c2)),
            SchoolClass('sat c3', cron('sat', c3)),
        ])

    def test_schedule_get_current__when_current_has_impossible_extra_condition(self):
        moment = arrow.get('2021-07-01T08:30:00', tzinfo='Europe/Moscow').datetime

        activities = self.schedule.get_current(moment)

        self.assertEqual(activities, ())

    def test_schedule_get_current__when_many_current_and_one_has_impossible_extra_condition(self):
        moment = arrow.get('2021-07-02T08:30:00', tzinfo='Europe/Moscow').datetime

        activities = self.schedule.get_current(moment)

        self.assertEqual(len(activities), 1)
        self.assertEqual(activities[0].payload, 'fri c12')
        self.assertEqual(
            activities[0].start,
            arrow.get('2021-07-02T08:00:00', tzinfo='Europe/Moscow').datetime
        )
        self.assertEqual(
            activities[0].end,
            arrow.get('2021-07-02T09:35:00', tzinfo='Europe/Moscow').datetime
        )

    def test_schedule_get_next__when_has_impossible_extra_condition(self):
        moment = arrow.get('2021-07-01T07:30:00', tzinfo='Europe/Moscow').datetime

        activities = self.schedule.get_next(moment)

        self.assertEqual(len(activities), 1)
        self.assertEqual(activities[0].payload, 'thu c2')
        self.assertEqual(
            activities[0].start,
            arrow.get('2021-07-01T09:50:00', tzinfo='Europe/Moscow').datetime
        )
        self.assertEqual(
            activities[0].end,
            arrow.get('2021-07-01T11:25:00', tzinfo='Europe/Moscow').datetime
        )

    def test_schedule_get_next__when_many_next_and_one_has_impossible_extra_condition(self):
        moment = arrow.get('2021-07-02T07:30:00', tzinfo='Europe/Moscow').datetime

        activities = self.schedule.get_next(moment)

        self.assertEqual(len(activities), 1)
        self.assertEqual(activities[0].payload, 'fri c12')
        self.assertEqual(
            activities[0].start,
            arrow.get('2021-07-02T08:00:00', tzinfo='Europe/Moscow').datetime
        )
        self.assertEqual(
            activities[0].end,
            arrow.get('2021-07-02T09:35:00', tzinfo='Europe/Moscow').datetime
        )

    def test_schedule_get_current_or_next__when_has_impossible_extra_condition(self):
        moment = arrow.get('2021-07-02T08:30:00', tzinfo='Europe/Moscow').datetime

        activities, is_current = self.schedule.get_current_or_next(
            moment, return_is_current=True
        )

        self.assertTrue(is_current)
        self.assertEqual(len(activities), 1)
        self.assertEqual(activities[0].payload, 'fri c12')
        self.assertEqual(
            activities[0].start,
            arrow.get('2021-07-02T08:00:00', tzinfo='Europe/Moscow').datetime
        )
        self.assertEqual(
            activities[0].end,
            arrow.get('2021-07-02T09:35:00', tzinfo='Europe/Moscow').datetime
        )


class TestScheduleWithNeverStartingActivities(unittest.TestCase):
    def setUp(self):
        self.schedule = ManualSchedule([
            NeverStartingPeriodicActivity('never 1'),
            NeverStartingPeriodicActivity('never 2'),
        ])

    def test_schedule_get_current__when_no_activities_ever_start(self):
        moment = arrow.get('2021-07-01T08:30:00', tzinfo='Europe/Moscow').datetime

        activities = self.schedule.get_current(moment)

        self.assertEqual(activities, ())

    def test_schedule_get_next__when_no_activities_ever_start(self):
        moment = arrow.get('2021-07-01T08:30:00', tzinfo='Europe/Moscow').datetime

        activities = self.schedule.get_next(moment)

        self.assertEqual(activities, ())

    def test_schedule_get_current_or_next__when_no_activities_ever_start(self):
        moment = arrow.get('2021-07-01T08:30:00', tzinfo='Europe/Moscow').datetime

        activities, is_current = self.schedule.get_current_or_next(
            moment, return_is_current=True
        )

        self.assertEqual(activities, ())
        self.assertIsNone(is_current)


if __name__ == '__main__':
    unittest.main()



from abc import (
    ABC,
    abstractmethod,
)
from datetime import (
    datetime,
)
from typing import (
    Any,
    Optional,
    Tuple,
    Union,
)


seconds = float


class StaticTimeSpanABC(ABC):
    start: Optional[datetime] = None
    end: Optional[datetime] = None

    @classmethod
    def __subclasshook__(cls, subclass):
        return (
            hasattr(subclass, 'start') and
            hasattr(subclass, 'end')
        )

    def __str__(self):
        return f'[{self.start!s} ... {self.end!s}]'


class Instrumented_StaticTimeSpanABC(StaticTimeSpanABC):
    @abstractmethod
    def __bool__(self) -> bool:
        pass
    
    @abstractmethod
    def __hash__(self) -> int:
        pass

    @abstractmethod
    def __eq__(self, other: Any) -> bool:
        pass

    @abstractmethod
    def __lt__(self, other: StaticTimeSpanABC) -> bool:
        """TimeSpan with no start is considered to be
        infinitely far in the future, so it's never less than any other one
        """

    @abstractmethod
    def __gt__(self, other: StaticTimeSpanABC) -> bool:
        """TimeSpan with no start is considered to be
        infinitely far in the future, so it's always greater than any other one
        """


class PeriodicTimePointABC(ABC):
    @abstractmethod
    def get_next(self, moment: datetime) -> Optional[datetime]:
        """Return value must be greater than ``moment``"""


class PeriodicTimeSpanABC(ABC):
    @abstractmethod
    def is_ongoing(self, moment: datetime) -> bool:
        pass

    @abstractmethod
    def get_current(self, moment: datetime) -> Instrumented_StaticTimeSpanABC:
        """Returns TimeSpan that either contains ``moment`` or is falsy"""

    @abstractmethod
    def get_next(self, moment: datetime) -> Instrumented_StaticTimeSpanABC:
        """Returns the earliest TimeSpan that starts after ``moment``;
        Returned TimeSpan may be falsy if there is no next period;
        Return value start time must be greater than ``moment``
        """

    @abstractmethod
    def get_current_or_next(
        self, moment: datetime, *, return_is_current: bool = False
    ) -> Union[Instrumented_StaticTimeSpanABC, Tuple[Instrumented_StaticTimeSpanABC, Optional[bool]]]:
        """Tries to return TimeSpan that contains ``moment``;
        If that is falsy (which means there is no ongoing period),
        returns the earliest TimeSpan that starts after ``moment`` which may be falsy too;
        Falsy TimeSpan return value means there is no next period as well
        """


class PeriodicActivityABC(PeriodicTimeSpanABC):
    payload: Any = None



from setuptools import setup
from pathlib import Path


README = Path("README.md").read_text()


setup(
    name='schedpack',
    packages=['schedpack'],
    version='1.1.1',
    license='MIT',
    description='Package for scheduling activities that last some time',
    long_description=README,
    long_description_content_type='text/markdown',
    author='borisoid',
    url='https://github.com/Borisoid/schedpack',
    keywords=['schedule', 'lasting', 'cron'],
    install_requires=[
        'croniter==1.0.15',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
    ],
)
